/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "my-math.h"
#include "my-stdlib.h"
#include "my-string.h"
#include "my-time.h"

#include "config.h"
#include "functions.h"
#include "log.h"
#include "random.h"
#include "storage.h"
#include "utils.h"

static int
parse_number(const char *str, int *result)
{
    char	*p;
    int		negative = 0;
    
    while (*str  &&  *str == ' ')
	str++;
    if (*str == '-') {
	str++;
	negative = 1;
    }
    *result = strtol(str, &p, 10);
    if (p == str)
	return 0;
    while (*p) {
	if (*p != ' ')
	    return 0;
	p++;
    }
    if (negative)
	*result = -*result;
    return 1;
}

static int
parse_object(const char *str, Objid *result)
{
    int		number;
    
    while (*str  &&  *str == ' ')
	str++;
    if (*str == '#')
	str++;
    if (parse_number(str, &number)) {
	*result = number;
	return 1;
    } else
	return 0;
}

enum error
become_number(Var in, int *ret, char called_from_tonum)
{ 
    switch (in.type) {
      case TYPE_NUM:
	*ret = in.v.num;
	break;
      case TYPE_STR: 
	if (!(called_from_tonum
	      ? parse_number(in.v.str, ret)
	      : parse_object(in.v.str, ret)))
#if I_HAD_MY_DRUTHERS
	    return E_INVARG;
#else
	*ret = 0;
#endif
	break;
      case TYPE_OBJ:
	*ret = in.v.obj;
	break;
      case TYPE_ERR:
	*ret = in.v.err;
	break;
      case TYPE_LIST:
	return E_TYPE;
      default:
	errlogf("BECOME_NUMBER: Impossible var type: %d\n", (int)in.type);
    }
    return E_NONE;
}

/**** built in functions ****/

static package
bf_tonum(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r; 
    enum error e;
    
    r.type = TYPE_NUM;
    e = become_number(arglist.v.list[1], &(r.v.num), 1);
    
    free_var(arglist);
    if (e != E_NONE)
	return make_error_pack(e);
    
    return make_var_pack(r);
}

static package
bf_min(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    int		min = arglist.v.list[1].v.num;
    int		i;
    
    r.type = TYPE_NUM;
    for (i = 2; i <= arglist.v.list[0].v.num; i++) {
	if (arglist.v.list[i].type != TYPE_NUM) {
	    free_var(arglist);
	    return make_error_pack(E_TYPE);
	} else if (arglist.v.list[i].v.num < min)
	    min = arglist.v.list[i].v.num;
    }
    r.v.num = min;
    
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_max(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    int		max = arglist.v.list[1].v.num;
    int		i;
    
    r.type = TYPE_NUM;
    for (i = 2; i <= arglist.v.list[0].v.num; i++) {
	if (arglist.v.list[i].type != TYPE_NUM) {
	    free_var(arglist);
	    return make_error_pack(E_TYPE);
	} else if (arglist.v.list[i].v.num > max)
	    max = arglist.v.list[i].v.num;
    }
    r.v.num = max;
    
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_abs(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    r.type = TYPE_NUM;
    if (arglist.v.list[1].v.num < 0)
	r.v.num = -arglist.v.list[1].v.num;
    else
	r.v.num = arglist.v.list[1].v.num;
    free_var(arglist);
    return make_var_pack(r);
}              

static package
bf_sqrt(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;

    if (arglist.v.list[1].v.num < 0) {
	free_var(arglist);
	return make_error_pack(E_INVARG);
    }
    r.type = TYPE_NUM;
    r.v.num = (int)sqrt((double)arglist.v.list[1].v.num);
    free_var(arglist);
    return make_var_pack(r);
}
    
static package
bf_time(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    r.type = TYPE_NUM;
    r.v.num = time(0);
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_ctime(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    time_t	c;
    char	buffer[50];
    
    if (arglist.v.list[0].v.num == 1) {
	c = arglist.v.list[1].v.num;
    } else {
	c = time(0);
    }

    { /* Format the time, including a timezone name */
#if HAVE_STRFTIME
	strftime(buffer, 50, "%a %b %d %H:%M:%S %Y %Z", localtime(&c));
#else
#  if HAVE_TM_ZONE
	struct tm      *t = localtime(&c);
	char	       *tzname = t->tm_zone;
#  else
#    if !HAVE_TZNAME
	const char     *tzname = "XXX";
#    endif
#  endif	
	
	strcpy(buffer, ctime(&c));
	buffer[24] = ' ';
	strncpy(buffer + 25, tzname, 3);
	buffer[28] = '\0';
#endif
    }
    
    if (buffer[8] == '0')
	buffer[8] = ' ';
    r.type = TYPE_STR;
    r.v.str = str_dup(buffer);
    
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_random(Var arglist, Byte next, void *vdata, Objid progr)
{
    int num = arglist.v.list[1].v.num;
    
    free_var(arglist);
    
    if (num <= 0) 
	return make_error_pack(E_INVARG);
    else {
	Var r; 

	r.type = TYPE_NUM;
	r.v.num = RANDOM() % num + 1;
	return make_var_pack(r);
    }
}

void
register_numbers(void)
{ 
    (void) register_function("tonum", 1, 1, bf_tonum, TYPE_ANY);
    (void) register_function("min", 1, -1, bf_min, TYPE_NUM);
    (void) register_function("max", 1, -1, bf_max, TYPE_NUM);
    (void) register_function("abs", 1, 1, bf_abs, TYPE_NUM);
    (void) register_function("random", 1, 1, bf_random, TYPE_NUM);
    (void) register_function("sqrt", 1, 1, bf_sqrt, TYPE_NUM);
    (void) register_function("time", 0, 0, bf_time);
    (void) register_function("ctime", 0, 1, bf_ctime, TYPE_NUM);
}

char rcsid_numbers[] = "$Id: numbers.c,v 1.9 1992/10/23 23:03:47 pavel Exp $";

/* $Log: numbers.c,v $
 * Revision 1.9  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.8  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.7  1992/10/17  20:47:26  pavel
 * Global rename of strdup->str_dup, strref->str_ref, vardup->var_dup, and
 * varref->var_ref.
 *
 * Revision 1.6  1992/09/26  18:02:49  pavel
 * Fixed bug whereby passing negative numbers to random() failed to evoke
 * E_INVARG.
 *
 * Revision 1.5  1992/09/14  17:31:52  pjames
 * Updated #includes.
 *
 * Revision 1.4  1992/09/08  22:01:42  pjames
 * Renamed bf_num.c to numbers.c.  Added `become_number()' from bf_type.c
 *
 * Revision 1.3  1992/08/10  17:36:26  pjames
 * Updated #includes.  Used new regisration method.  Add bf_sqrt();
 *
 * Revision 1.2  1992/07/20  23:51:47  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 */
